/***************************************************************************
 *   Copyright (C) 2003 by Roberto Natella                                 *
 *   roberto.natella@virgilio.it                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <klocale.h>
#include <kconfig.h>
#include <kapplication.h>
#include <kmessagebox.h>
#include <kaboutapplication.h>
#include <kstandarddirs.h>
#include <kdebug.h>
#include <kstandarddirs.h>
#include <krun.h>
#include <ktempfile.h>
#include <kmimetype.h>
#include <kpopupmenu.h>
#include <kprocess.h>
#include <kprocio.h>
#include <kio/job.h>
#include <kio/jobclasses.h>
#include <kaudioplayer.h>
#include <dcopclient.h>
#include <kprotocolmanager.h>
#include <kiconloader.h>

#include <qtooltip.h>
#include <qstringlist.h>
#include <qdom.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qregexp.h>
#include <qdatetime.h>
#include <qcursor.h>

#include "komics.h"
#include "pref.h"

const uint defaultTimeout = 30;  // half an hour

Komics::Komics( QWidget * parent, KConfig * config, bool isApplet ) : DCOPObject("Komics"), KomicsWidget( parent )
{
    // Get the current application configuration handle
    m_kconfig = config;

    m_applet = isApplet;

    connect( this, SIGNAL(leftclickedSignal(QMouseEvent*)),
             this, SLOT(leftclicked(QMouseEvent*)) );
    connect( this, SIGNAL(rightclickedSignal(QMouseEvent*)),
             this, SLOT(rightclicked(QMouseEvent*)) );
    connect( kapp, SIGNAL(shutDown()),
             this, SLOT(shutDown()) );

    if( !loadConfig() )
    {
        kdDebug() << "Unable to read configuration files.\n";
    }

    if( !readHistory() )
    {
        kdDebug() << "Unable to read history files.\n";
    }

    m_prefopened = false;
    m_killing = false;
    m_hasplayed = false;    // If true, advising sound isn't played
    m_id = 0;

    m_kconfig->setGroup("General Setup");
    QString onstart = m_kconfig->readEntry("fetch_on_start", "true");
    QString autodownload = m_kconfig->readEntry("autodownload", "false");
    QString refresh = m_kconfig->readEntry("refresh", "0");
    
    m_sound = m_kconfig->readEntry("sound", "");

    m_date = QDate::currentDate();
    
    bool ok;
    m_refresh = refresh.toInt(&ok);
    if( !ok || m_refresh <= 0 )
    {
        m_refresh = defaultTimeout;
    }

    if( !QFile::exists(m_sound) )
    {
        m_sound = "";
    }
    
    m_onstart = (onstart == "true");
    m_autodownload = (autodownload == "true");

    if( m_onstart )
    {
        poll();
    }

    m_timer = new QTimer(this,"pollTimer");
    connect( m_timer, SIGNAL(timeout()), this, SLOT(poll()) );
    m_timer->start(m_refresh * 60 * 1000);

    m_itemId = 0;
    m_popup = new KPopupMenu(0, "rmb");
    fillPopupMenu();
    m_popup->adjustSize();

    m_client = new DCOPClient;
    if( !m_client->isApplicationRegistered("komics") )
    {
        m_client->attach();
        m_client->registerAs("komics", false);
    }
}

Komics::~Komics()
{
    shutDown();
    m_client->detach();

    delete m_popup;
    delete m_timer;
    delete m_client;
}

void Komics::about()
{
    KAboutData about ( "komics1", I18N_NOOP("Komics"), VERSION,
                       I18N_NOOP(__comment), KAboutData::License_GPL_V2,
                       "(C) 2003 Natella Roberto", 0, "roberto.natella@virgilio.it" );

    about.addAuthor("Roberto Natella", 0,
                    "roberto.natella@virgilio.it", "http://www.orson.it/" );

    KAboutApplication a(&about);
    a.exec();
}

/*
void Komics::help()
{
    KMessageBox::information(0, i18n("This is a help box"));
}
*/

void Komics::preferences()
{
    if( m_prefopened )
        return;

    m_prefopened = true;

    KomicsPreferences dialog(this);

    //dialog.stripsPage()->m_stripsdialog->setContents(m_strips, m_criteria);
    dialog.setContents(m_strips, m_criteria);

    dialog.setupPage()->setOnStart(m_onstart);
    dialog.setupPage()->setAutodownload(m_autodownload);
    dialog.setupPage()->setRefresh(m_refresh);
    dialog.setupPage()->setSound(m_sound);
    
    dialog.exec();

    m_prefopened = false;
}

void Komics::fillPopupMenu()
{
    m_popup->clear();

    m_popup->insertTitle(SmallIcon("komics1"), "Komics");

    m_itemId = 0;

    QMap< QString, QMap<QString, QString> >::const_iterator it;

    for( it = m_strips.begin(); it != m_strips.end(); it++ )
    {
        if( (*it)["activated"] == "1" )
            m_popup->insertItem(it.key(), m_itemId++);
    }

    if( !m_applet )
    {
        m_popup->insertSeparator();
        
        m_idAbout     = m_popup->insertItem( SmallIcon("clock"), i18n("&About Komics") );
        m_idConfigure = m_popup->insertItem( SmallIcon("configure"), i18n("&Configure Komics...") );
        m_idQuit      = m_popup->insertItem( SmallIcon("exit"), i18n("&Quit Komics") );
    }
}

bool Komics::loadConfig()
{
    m_strips.clear();
    m_criteria.clear();
    m_stripsStack.clear();
    m_stripsServed.clear();

    // If we found a config file in ~/.kde dir, we give him priority;
    // else, we read all others
    
    QStringList resources = kapp->dirs()->saveLocation("data", "komics/komics.xml", false);

    if( !QFileInfo(resources[0]).exists() )
    {
        resources = kapp->dirs()->findAllResources("data", "komics/komics.xml", true);
        
        if( resources.size() == 0 )
        {
            kdDebug() << "No XML resource file found.\n";
            return false;
        }
    }

    for( uint i = 0; i < resources.size(); i++ )
    {
        QDomDocument doc("resource");
        QFile file(resources[i]);
        doc.setContent( &file );

        QDomElement docElement = doc.documentElement();

        QDomNodeList stripsList = doc.elementsByTagName(QString("strip"));

        if( stripsList.length() == 0 )
        {
            kdDebug() << "No <strip> tag in " << resources[i] << "\n";
            return false;
        }

        for( uint n = 0; n < stripsList.count(); n++ )
        {
            QString name =
                   stripsList.item(n).attributes().namedItem("name").nodeValue();

            if( name.isEmpty() )
            {
                kdDebug() << "Missing \"name\" attribute in <strip> tag in"\
                          << resources[i] << "\n";
                return false;
            }

            QDomElement strip = stripsList.item(n).toElement();

            QDomNodeList execList = strip.elementsByTagName(QString("executable"));
            QDomNodeList urlList = strip.elementsByTagName(QString("url"));
            QDomNodeList formatList = strip.elementsByTagName(QString("format"));
            QDomNodeList referrerList = strip.elementsByTagName(QString("referrer"));
            QDomNodeList activatedList = strip.elementsByTagName(QString("activated"));
            QString executable;
            QString url;
            QString format;
            QString referrer;
            QString activated;

            if( execList.length() > 0 )
            {
                executable = execList.item(0).firstChild().nodeValue();
            }

            if( formatList.length() > 0 )
            {
                format = formatList.item(0).firstChild().nodeValue();
            }

            if( referrerList.length() > 0 )
            {
                referrer = referrerList.item(0).firstChild().nodeValue();
            }

            if( activatedList.length() > 0 )
            {
                activated = activatedList.item(0).firstChild().nodeValue();
            }

            if( urlList.length() > 0 )
            {
                url = urlList.item(0).firstChild().nodeValue();
            }
            else
            {
                kdDebug() << "Tag <url> missing in " << resources[i] << "\n";
                return false;
            }

            QDomNodeList criteriaList = strip.elementsByTagName(QString("criteria"));

            if( criteriaList.length() == 0 )
            {
                kdDebug() << "No <criteria> tag in " << resources[i] << "\n";
                return false;
            }

            for( uint j = 0; j < criteriaList.length(); j++ )
            {

                QString inverse =
                    criteriaList.item(j).attributes().namedItem("inverse").nodeValue();

                int indications = 0;

                QDomElement criteria = criteriaList.item(j).toElement();

                QDomNodeList days = criteria.elementsByTagName(QString("day"));
                int day = -1;
                bool ok = false;

                if( days.length() > 0 )
                {
                    day = days.item(0).firstChild().nodeValue().toInt(&ok);
                    if( ok && day > 0 && day < 32 )
                    {
                        indications++;
                    }
                    else
                    {
                        day = -1;
                    }
                }

                QDomNodeList months = criteria.elementsByTagName(QString("month"));
                int month = -1;
                ok = false;

                if( months.length() > 0 )
                {
                    month = months.item(0).firstChild().nodeValue().toInt(&ok);
                    if( ok && month > 0 && month < 13 )
                    {
                        indications++;
                    }
                    else
                    {
                        month = -1;
                    }
                }

                QDomNodeList hours = criteria.elementsByTagName(QString("hour"));
                int hour = -1;
                ok = false;

                if( hours.length() > 0 )
                {
                    hour = hours.item(0).firstChild().nodeValue().toInt(&ok);
                    if( ok && hour >= 0 && hour < 24 )
                    {
                        indications++;
                    }
                    else
                    {
                        hour = -1;
                    }
                }

                QDomNodeList minutes = criteria.elementsByTagName(QString("minute"));
                int minute = -1;
                ok = false;

                if( minutes.length() > 0 )
                {
                    minute = minutes.item(0).firstChild().nodeValue().toInt(&ok);
                    if( ok && minute >= 0 && minute < 60 )
                    {
                        indications++;
                    }
                    else
                    {
                        minute = -1;
                    }
                }

                QDomNodeList daynames = criteria.elementsByTagName(QString("dayname"));
                int dayname = -1;
                ok = false;

                if( daynames.length() > 0 )
                {
                    dayname = daynames.item(0).firstChild().nodeValue().toInt(&ok);

                    if( ok && dayname > 0 && dayname < 8 )
                    {
                        indications++;
                    }
                    else
                    {
                        dayname = -1;
                    }
                }

                if( indications == 0 )
                {
                    kdDebug() << "<criteria> tags must at least contain " <<
                                 "a valid indication.\n";
                    return false;
                }

                m_strips[name]["url"] = url;
                m_strips[name]["executable"] = executable;
                m_strips[name]["format"] = format;
                m_strips[name]["referrer"] = referrer;
                m_strips[name]["activated"] = (activated == "0" ? "0" : "1");

                QMap<QString,int> tmp;
                tmp["day"] = day;
                tmp["month"] = month;
                tmp["hour"] = hour;
                tmp["minute"] = minute;
                tmp["dayname"] = dayname;
                tmp["inverse"] = (inverse == "1" ? 1 : -1);

                m_criteria[name].push_back(tmp);

            }
        }

    }

    return true;
}

bool Komics::writeConfig(const StripsMap& strips, const CriteriaMap& criteria)
{
    
    QString savepath = kapp->dirs()->saveLocation("data", "komics/", true);
    savepath += "komics.xml";
    
    QFile file( savepath );

    QDomDocument doc("Komics");

    QDomNode root = doc.createElement(QString("xml"));
    doc.appendChild(root);

    QMap< QString, QMap<QString, QString> >::const_iterator it;

    for( it = strips.begin(); it != strips.end(); it++ )
    {
        if( it.key().isEmpty() ) continue;
    
        QDomElement node, subNode, criteriaNode;

        node = doc.createElement(QString("strip"));

        node.setAttribute("name", it.key());

        subNode = doc.createElement(QString("url"));

        QString url = it.data()["url"];

        if( url.isEmpty() )
        {
            KMessageBox::sorry(0, i18n( QString("No URL set for %1").arg(it.key()) ));
            return false;
        }

        subNode.appendChild( doc.createTextNode(url) );
        node.appendChild(subNode);

        QString executable = it.data()["executable"];

        if( !executable.isEmpty() )
        {
            subNode = doc.createElement(QString("executable"));
            subNode.appendChild( doc.createTextNode(executable) );
            node.appendChild(subNode);
        }

        QString format = it.data()["format"];

        if( !format.isEmpty() )
        {
            subNode = doc.createElement(QString("format"));
            subNode.appendChild( doc.createTextNode(format) );
            node.appendChild(subNode);
        }

        QString referrer = it.data()["referrer"];

        if( !referrer.isEmpty() )
        {
            subNode = doc.createElement(QString("referrer"));
            subNode.appendChild( doc.createTextNode(referrer) );
            node.appendChild(subNode);
        }

        QString activated = it.data()["activated"];

        if( !activated.isEmpty() )
        {
            if( activated != "0" && activated != "1" )
                activated = "1";

            subNode = doc.createElement(QString("activated"));
            subNode.appendChild( doc.createTextNode(activated) );
            node.appendChild(subNode);
        }

        if( criteria[it.key()].begin() == criteria[it.key()].end() )
        {
            KMessageBox::sorry(0,
                               i18n( QString("No criteria set for %1").arg(it.key()) ));
            return false;
        }

        QValueVector< QMap< QString,int > >::const_iterator crit;
        for( crit = criteria[it.key()].begin();
             crit != criteria[it.key()].end(); crit++ )
        {
            criteriaNode = doc.createElement("criteria");

            if( (*crit)["inverse"] > 0 )
            {
                criteriaNode.setAttribute("inverse","1");
            }

            int day = (*crit)["day"];
            int month = (*crit)["month"];
            int hour = (*crit)["hour"];
            int minute = (*crit)["minute"];
            int dayname = (*crit)["dayname"];

            int indications = 0;

            if( day > 0 )
            {
                subNode = doc.createElement("day");
                subNode.appendChild( doc.createTextNode(QString("%1").arg(day)) );
                criteriaNode.appendChild(subNode);
                indications++;
            }

            if( month > 0 )
            {
                subNode = doc.createElement("month");
                subNode.appendChild( doc.createTextNode(QString("%1").arg(month)) );
                criteriaNode.appendChild(subNode);
                indications++;
            }

            if( hour >= 0 )
            {
                subNode = doc.createElement("hour");
                subNode.appendChild( doc.createTextNode(QString("%1").arg(hour)) );
                criteriaNode.appendChild(subNode);
                indications++;
            }

            if( minute >= 0 )
            {
                subNode = doc.createElement("minute");
                subNode.appendChild( doc.createTextNode(QString("%1").arg(minute)) );
                criteriaNode.appendChild(subNode);
                indications++;
            }

            if( dayname > 0 )
            {
                subNode = doc.createElement("dayname");
                subNode.appendChild( doc.createTextNode(QString("%1").arg(dayname)) );
                criteriaNode.appendChild(subNode);
                indications++;
            }

            if( indications == 0 )
            {
                KMessageBox::sorry(0,
                             i18n("Criteria must contain at least a valid indication"));
                return false;
            }

            node.appendChild(criteriaNode);
        }

        root.appendChild(node);
    }

    if( file.open( IO_WriteOnly | IO_Truncate ) )
    {
        QTextStream stream( &file );
        stream << doc.toString();
        file.close();
    }
    else
    {
        KMessageBox::sorry(0,
                        i18n( QString("Unable to write to %1").arg(savepath) ));
        return false;
    }

    return true;
}

bool Komics::poll()
{
    QDate date = QDate::currentDate();
    QTime time = QTime::currentTime();

    int day = date.day();
    int month = date.month();
    int hour = time.hour();
    int minute = time.minute();
    int dayname = date.dayOfWeek();

    // If midnight has passed, we clear the history of the previous day
    if( date != m_date )
    {
        flushHistory();
        m_date = date;
    }
    
    QMap< QString, QValueVector< QMap< QString,int > > >::const_iterator strip;
    QValueVector< QMap< QString,int > >::const_iterator item;

    bool pushed = false;

    for( strip = m_criteria.begin(); strip != m_criteria.end(); strip++ )
    {
        QString name = strip.key();

        if( m_strips[name]["activated"] == "0" )
            continue;

        int i = 0;

        for( item = strip.data().begin(); item != strip.data().end(); item++ )
        {
            bool invert = ((*item)["inverse"] == 1);

            if(    (*item)["day"] < 0
                || ((*item)["day"] == day && !invert)
                || ((*item)["day"] != day && invert) )
            {
              if(    (*item)["month"] < 0
                  || ((*item)["month"] == month && !invert)
                  || ((*item)["month"] != month && invert) )
              {
                if(    (*item)["dayname"] < 0
                    || ((*item)["dayname"] == dayname && !invert)
                    || ((*item)["dayname"] != dayname && invert) )
                {
                  if(    ((*item)["hour"] < 0)
                      || (( ((*item)["hour"] < hour)
                      || (((*item)["hour"] == hour) && ((*item)["minute"] <= minute)) )
                      && !invert)
                      || (( ((*item)["hour"] > hour)
                      || (((*item)["hour"] == hour) && ((*item)["minute"] >= minute)) )
                      && invert)
                    )
                  {
                    if( qFind( m_stripsServed.begin(),m_stripsServed.end(),
                        strip.key()+QString("_%1").arg(i) ) == m_stripsServed.end() )
                    {
                      m_stripsStack.push_back(name+QString("_%1").arg(i));
                      if( !pushed )
                      {
                        pushed = true;
                        setIcon(Ready);
                      }
                    }
                  }
                }
              }
            }

            i++;
        }
    }
    
    if( pushed && m_autodownload )
    {
        leftclicked(NULL);
    }

    if( pushed && !m_hasplayed && !m_sound.isEmpty() )
    {
        KAudioPlayer::play(m_sound);
        m_hasplayed = true;
    }

    return pushed;
}

void Komics::rightclicked(QMouseEvent * /*e*/)
{
    QSize size = m_popup->sizeHint();
    QPoint cursor = QCursor::pos();
    int item = m_strips.count();

    if ( size.height() < cursor.y() )
    {
        item = m_popup->exec(QPoint(cursor.x(), cursor.y() - size.height()));
    }
    else
    {
        item = m_popup->exec(QPoint(cursor.x(), cursor.y()));
    }

    if( item == (int)m_strips.count() )
        return;

    QMap< QString, QMap<QString, QString> >::const_iterator it = m_strips.begin();

    if( item >= 0 )
    {
        while( --item >= 0 )
            it++;

        setIcon(Loading);
        showStrip(it.key());
    }
    else
    {
        if( item == m_idAbout )
            about();
        else if ( item == m_idConfigure )
            preferences();
        else if ( item == m_idQuit )
            kapp->quit();
    }
}

void Komics::leftclicked(QMouseEvent * /*e*/)
{
    if( state() == Loading )
    {
        m_killing = true;
        
        killall();
        checkIfResetIcon();
        
        m_killing = false;
        
        return;
    }

    if( state() != Ready || m_stripsStack.size() == 0 )
    {
        checkIfResetIcon();
        return;
    }

    m_hasplayed = false;

    setIcon(Loading);

    QValueVector<QString> processed;

    int size = m_stripsStack.size();
    
    for( uint n = size; n > 0; n-- )
    {
        m_stripsServed.push_back( m_stripsStack[n-1] );
        m_stripsStack.pop_back();

        QRegExp re( "^(.*)_\\d+$", true, false );
        int pos = re.search(m_stripsStack[n-1]);

        if( pos > -1 )
        {
            QString strip = re.cap(1);
            if( qFind(processed.begin(),processed.end(),strip) == processed.end() )
            {
                processed.push_back(strip);
            }
        }
    }

    QValueVector<QString>::const_iterator it;
    for( it = processed.begin(); it != processed.end(); it++ )
        showStrip(*it);
    
}

void Komics::showStrip( const QString & strip )
{
    QString url = m_strips[strip]["url"];
    QString command = m_strips[strip]["executable"];

    KURL tmpurl( url );

    if( tmpurl.isValid() && !tmpurl.isLocalFile() )
    {
        url = tmpurl.url();
    }
    else
    {
        QString execute;

        if( tmpurl.isValid() )
        {
            // it's a local file
            execute = tmpurl.filename();
        }
        else
        {
            // the url was malformed
            execute = url;
        }

        if( getCommand(execute) )
        {
            KProcess * process = new KProcess();

            m_processes[process]["output"] = "";
            m_processes[process]["strip"] = strip;

            connect( process, SIGNAL(receivedStdout(KProcess*,char*,int)),
                     this, SLOT(receivedStdout(KProcess*,char*,int)) );
            connect( process, SIGNAL(processExited(KProcess*)),
                     this, SLOT(processExited(KProcess*)) );

            if( KProtocolManager::useProxy() )
            {
                QString proxy;

                // deprecated

                proxy = KProtocolManager::proxyFor("http");
                process->setEnvironment("KOMICS_HTTP_PROXY", proxy);

                proxy = KProtocolManager::proxyFor("https");
                process->setEnvironment("KOMICS_HTTPS_PROXY", proxy);

                proxy = KProtocolManager::proxyFor("ftp");
                process->setEnvironment("KOMICS_FTP_PROXY", proxy);
            }
            
            (*process) << execute;
            if( !process->start( KProcIO::NotifyOnExit, KProcIO::Stdout ) )
            {
                if(!m_killing)
                {
                    KMessageBox::sorry(0,i18n("Unable to execute '%1'.").arg(execute));
                    m_processes.erase(process);
                    delete process;
                }
                checkIfResetIcon();
            }
        }

        return;
    }

    downloadStart(url,command,strip);
}

bool Komics::runCommand( const QString & command, const QString & url )
{
    checkIfResetIcon();

    QString cmd = command;

    if( cmd.isEmpty() )
    {
        QString mime = KMimeType::findByURL(url)->name();
        return KRun::runURL( url, mime );
    }
    else
    {
        subst( cmd, "%u", url );
        return KRun::runCommand(cmd);
    }
}

void Komics::receivedStdout( KProcess * proc, char * buffer, int len )
{
    if(m_killing)
        return;
        
    buffer[len] = 0;
    m_processes[proc]["output"] += QString(buffer);
}

void Komics::processExited( KProcess * proc )
{
    if(m_killing)
        return;

    QStringList list = QStringList::split('\n', m_processes[proc]["output"], false);

    QString strip = m_processes[proc]["strip"];

    m_processes.erase(proc);
    delete proc;

    if( list.count() == 0 )
    {
        checkIfResetIcon();
        return;
    }
    
    for( uint i=0; i < list.count(); i++ )
    {
        downloadStart( list[i], m_strips[strip]["executable"], strip, i );
    }
}

bool Komics::getCommand( QString & command )
{
    QRegExp re("([^\\./][^/]+)$");

    if( re.search(command) < 0 )
    {
        KMessageBox::sorry(0,i18n("URL '%1' is malformed.").arg(command) );
        return false;
    }

    QString tmp = re.cap(1);

    QStringList resources = kapp->dirs()->findDirs("data", "komics");

    uint n = 0;
    bool found = false;
    while( n<resources.size() && !found )
    {
        if( QFile::exists(resources[n]+tmp) )
        {
            found = true;
        }
        n++;
    }

    if( !found )
    {
        KMessageBox::sorry( 0, i18n("Cannot find '%1'.").arg(tmp) );
        checkIfResetIcon();
        return false;
    }

    command = resources[n-1]+tmp;

    return true;
}

bool Komics::downloadStart( const KURL &url, const QString &command, const QString &strip, uint number )
{
    QString extension;
    QRegExp re("(\\.[^\\.]+)$");
    re.setMinimal(true);

    if( re.search(url.fileName()) >= 0 )
    {
        extension = re.cap(1);
    }

    KTempFile * file = new KTempFile(QString::null,extension);

    KIO::Job *job = KIO::get(url, false, true);

    QString referrer = m_strips[strip]["referrer"];

    if( !referrer.isEmpty() )
    {
        job->addMetaData("referrer", referrer);
    }

    m_jobshandles[job] = file;
    m_jobsinfo[job]["command"] = command;
    m_jobsinfo[job]["url"] = url.url();
    m_jobsinfo[job]["strip"] = strip;
    
    if(number > 0) m_jobsinfo[job]["number"] = QString("%1").arg(number);

    connect( job, SIGNAL(data(KIO::Job *, const QByteArray &)),
             this, SLOT(downloadData(KIO::Job *, const QByteArray &)) );
    connect( job, SIGNAL(result(KIO::Job *)),
             this, SLOT(downloadResult(KIO::Job *)) );

    return true;
}

void Komics::downloadData( KIO::Job *job, const QByteArray &data )
{
    if( m_jobshandles[job] )
    {
        QFile * file = m_jobshandles[job]->file();
        file->writeBlock(data.data(), data.size());
    }
}

void Komics::downloadResult( KIO::Job *job )
{
    KTempFile * file = m_jobshandles[job];
    file->file()->close();

    if (job->error())
    {
        if(!m_killing)
            KMessageBox::sorry(0, i18n("Cannot download '%1'.").arg(m_jobsinfo[job]["strip"]));
        checkIfResetIcon();
    }
    else
    {
        QString path;
        QString strip = m_jobsinfo[job]["strip"];
        QString format = m_strips[strip]["format"];
        QString number = m_jobsinfo[job]["number"];

        if( !format.isEmpty() )
        {
            QDate date = QDate::currentDate();
            QTime time = QTime::currentTime();

            QString day = date.toString("dd");
            QString month = date.toString("MM");
            QString year = date.toString("yy");
            QString hour = time.toString("hh");
            QString minute = time.toString("mm");
            QString dayname = QDate::longDayName(date.dayOfWeek());

            path = format;

            subst( path, "%d", day );
            subst( path, "%m", month );
            subst( path, "%y", year );
            subst( path, "%h", hour );
            subst( path, "%t", minute );
            subst( path, "%n", dayname );

            if( !number.isEmpty() )
            {
                QFileInfo file = QFileInfo(path);

                QString dir = file.dirPath();
                QString name = file.baseName();
                QString ext = file.extension();

                path = dir + "/" + name + "-" + number;
                if( !ext.isEmpty() ) path += "." + ext;
            }

            KIO::file_copy(file->name(), path);
        }

        runCommand(m_jobsinfo[job]["command"], file->name());
    }

    m_jobshandles.erase( job );
    m_jobsinfo.erase( job );
    
    checkIfResetIcon();
}

void Komics::killall()
{
    QMap< KProcess*, QMap<QString,QString> >::const_iterator proc;

    for( proc = m_processes.begin(); proc != m_processes.end(); proc++ )
    {
        proc.key()->kill();
    }

    QMap<KIO::Job*, KTempFile *>::const_iterator job;

    for( job = m_jobshandles.begin(); job != m_jobshandles.end(); job++ )
    {
        job.key()->kill(false);
    }
}

void Komics::subst(QString &str, const QString &match, const QString &replace) const
{
    QRegExp re(match);
    int len = match.length();
    int pos = 0;
    while ( (pos = re.search(str,pos)) != -1 )
    {
        str = str.replace(pos,len,replace);
        pos += re.matchedLength();
    }
}

bool Komics::readHistory()
{
    m_stripsServed.clear();

    QStringList resources = kapp->dirs()->findAllResources("data",
                                                           "komics/komicshistory.xml",
                                                           true);

    if( resources.size() == 0 )
    {
        kdDebug() << "No XML history file found.\n";
        return false;
    }

    for( uint i = 0; i < resources.size(); i++ )
    {
        QDomDocument doc("history");
        QFile file(resources[i]);
        doc.setContent( &file );

        QDomElement docElement = doc.documentElement();

        QDomNodeList dayList = doc.elementsByTagName(QString("day"));
        QDomNodeList monthList = doc.elementsByTagName(QString("month"));
        int day;
        int month;

        if( dayList.length() > 0 )
        {
            bool ok = false;
            day = dayList.item(0).firstChild().nodeValue().toInt(&ok);
            if(!ok)
            {
                kdDebug() << "Malformed <day> tag in " << resources[i] << "\n";
                return false;
            }
        }
        else
        {
            kdDebug() << "No <day> tag found in " << resources[i] << "\n";
            return false;
        }

        if( monthList.length() > 0 )
        {
            bool ok = false;
            month = monthList.item(0).firstChild().nodeValue().toInt(&ok);
            if(!ok)
            {
                kdDebug() << "Malformed <month> tag in " << resources[i] << "\n";
                return false;
            }
        }
        else
        {
            kdDebug() << "No <month> tag found in " << resources[i] << "\n";
            return false;
        }

        QDate date = QDate::currentDate();

        int currentday = date.day();
        int currentmonth = date.month();

        if( currentday != day || currentmonth != month )
        {
            return true;
        }

        QDomNodeList criteriaList = doc.elementsByTagName(QString("criteria"));

        QRegExp re("(.+)_(\\d+)$");

        for( uint n = 0; n < criteriaList.count(); n++ )
        {
            QString criteria = criteriaList.item(n).firstChild().nodeValue();

            if( re.search(criteria) < 0 )
            {
                kdDebug() << "Malformed <criteria> tag in " << resources[i] << "\n";
                m_stripsServed.clear();
                return false;
            }

            QString strip = re.cap(1);
            QString num = re.cap(2);

            if( m_strips.find(strip) == m_strips.end() )
            {
                kdDebug() <<"Inexistent strip '" <<strip<<"' in "<<resources[i]<<"\n";
                m_stripsServed.clear();
                return false;
            }

            bool ok = false;
            uint id = num.toInt(&ok);
            if( !ok || id >= m_criteria[strip].size() )
            {
                kdDebug() <<"Invalid id '"<<num<<"' for '" <<strip<<"' in "<<resources[i]<<"\n";
                m_stripsServed.clear();
                return false;
            }

            m_stripsServed.push_back( criteria );
        }
    }

    return true;
}

bool Komics::writeHistory()
{
    QString historypath = kapp->dirs()->saveLocation("data", "komics/", true);
    historypath += "komicshistory.xml";
    
    QFile file( historypath );

    QDomDocument doc("history");

    QDomNode root = doc.createElement(QString("xml"));
    doc.appendChild(root);

    QDomElement node;

    QDate date = QDate::currentDate();

    QString day = QString("%1").arg(date.day());
    QString month = QString("%1").arg(date.month());

    node = doc.createElement(QString("day"));
    node.appendChild( doc.createTextNode(day) );
    root.appendChild(node);

    node = doc.createElement(QString("month"));
    node.appendChild( doc.createTextNode(month) );
    root.appendChild(node);

    for( uint n=0; n<m_stripsServed.size(); n++ )
    {
        node = doc.createElement(QString("criteria"));
        node.appendChild( doc.createTextNode(m_stripsServed[n]) );
        root.appendChild(node);
    }

    if( file.open( IO_WriteOnly | IO_Truncate ) )
    {
        QTextStream stream( &file );
        stream << doc.toString();
        file.close();
    }
    else
    {
        kdDebug() << "Unable to write history in " << historypath << "\n";
        return false;
    }

    return true;
}

void Komics::checkIfResetIcon()
{
    if( m_killing || m_processes.count()+m_jobshandles.count() == 0 )
    {
        if( m_stripsStack.size() == 0 )
        {
            setIcon(Waiting);
        }
        else
        {
            setIcon(Ready);
        }
    }
    else
    {
        setIcon(Loading);
    }
}

void Komics::flushHistory()
{
    m_stripsServed.clear();
}

// DCOP
bool Komics::download( const QString & strip )
{
    if( !m_strips.contains(strip) )
        return false;

    setIcon(Loading);
    showStrip(strip);

    return true;
}

// DCOP
QStringList Komics::stripsList()
{
    return QStringList( m_strips.keys() );
}

// DCOP
QString Komics::url( const QString & strip )
{
    if( !m_strips.contains(strip) )
        return QString("");

    QString url = m_strips[strip]["url"];

    KURL tmpurl( url );

    if( tmpurl.isValid() && !tmpurl.isLocalFile() )
    {
        url = tmpurl.url();
    }
    else
    {
        QString execute;

        if( tmpurl.isValid() )
        {
            // it's a local file
            execute = tmpurl.filename();
        }
        else
        {
            // the url was malformed
            execute = url;
        }
        
        getCommand(execute);
        
        url = execute;
    }
    
    return url;
}

bool Komics::lock()
{
    QString lock = locateLocal("data", "komics/lock");
    KSimpleConfig config(lock);

    int pid = config.readNumEntry("pid", -1);

    if ( pid > -1 )
    {
        KMessageBox::sorry(0, i18n( QString("Unable to start the application. Please check if another instance was already started. If not, delete the lockfile $HOME/.kde/share/apps/komics/lock .") ));
        return false;
    }

    config.writeEntry("pid", getpid());
    config.sync();

    return true;
}

bool Komics::unlock()
{
    QString lock = locateLocal("data", "komics/lock");
    KSimpleConfig config(lock);

    config.writeEntry("pid", -1);
    config.sync();

    return true;
}

void Komics::shutDown()
{
    writeHistory();
    unlock();
}

KomicsApplet::KomicsApplet(const QString& configFile, Type type, int actions, QWidget *parent, const char *name)
    : KPanelApplet(configFile, type, actions, parent, name)
{
    m_komics = new Komics(this, config(), true);
    m_komics->move(0,0);
    m_komics->show();
}

KomicsApplet::~KomicsApplet()
{
    delete m_komics;
}

void KomicsApplet::resizeEvent(QResizeEvent * e)
{
    m_komics->resizeEvent(e);
    KPanelApplet::resizeEvent(e);
}

int KomicsApplet::widthForHeight(int height) const
{
    return height;
}

int KomicsApplet::heightForWidth(int width) const
{
    return width;
}

void KomicsApplet::preferences()
{
    m_komics->preferences();
}

void KomicsApplet::about()
{
    m_komics->about();
}

extern "C"
{
    KPanelApplet* init( QWidget *parent, const QString configFile )
    {

        if(!Komics::lock())
        {
            return 0;
        }
        else
        {
            KGlobal::locale()->insertCatalogue("Komics");
            return new KomicsApplet(configFile, KPanelApplet::Normal,
                                    KPanelApplet::About | KPanelApplet::Preferences,
                                    parent, "Komics");
        }
    }
}
